# -*- coding: utf-8 -*-
"""Tests for the executor_easymotion."""

# Third Party Libraries
import pytest
from qtpy.QtCore import Qt


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos",
    [
        ("", [Qt.Key_Space, Qt.Key_Space, "w"], 0),
        ("a b.c d;e\n", [Qt.Key_Space, Qt.Key_Space, "w", "h"], 2),
        ("a b.c d;e\n", ["w", Qt.Key_Space, Qt.Key_Space, "w", "h"], 4),
        ("a b.c d;e\n", [Qt.Key_Space, Qt.Key_Space, "w", "e", "l"], 1),
        ("a b.c d;e\n", [Qt.Key_Space, Qt.Key_Space, "1", "w", "e"], 0),
        ('a b.c d"e\n\nf', [Qt.Key_Space, Qt.Key_Space, "w", "u"], 11),
        (
            "a\nb\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nf",
            [Qt.Key_Space, Qt.Key_Space, "w", "h"],
            2,
        ),
        (
            "1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1",
            [Qt.Key_Space, Qt.Key_Space, "w", "rh"],
            30,
        ),
    ],
)
def test_easymotion_w_cmd_in_normal(vim_bot, text, cmd_list, cursor_pos):
    """Test easymotion w command in normal."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert editor.textCursor().position() == cursor_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["l", "v", Qt.Key_Space, Qt.Key_Space, "w", "u"],
            26,
            [1, 27],
        )
    ],
)
def test_easymotion_w_cmd_in_visual(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test easymotion w command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["V", Qt.Key_Space, Qt.Key_Space, "w", "u"],
            26,
            [0, 50],
        )
    ],
)
def test_easymotion_w_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test easymotion w command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        (
            "a b.c de",
            ["y", Qt.Key_Space, Qt.Key_Space, "w", "l"],
            0,
            "a b.c de",
            '"',
            "a b.c ",
        ),
        (
            "a b.c de\nf k",
            ["d", Qt.Key_Space, Qt.Key_Space, "w", "u"],
            0,
            "k",
            '"',
            "a b.c de\nf ",
        ),
        (
            "a b.c de\nf k",
            ["c", Qt.Key_Space, Qt.Key_Space, "w", "u"],
            0,
            "k",
            '"',
            "a b.c de\nf ",
        ),
    ],
)
def test_easymotion_w_cmd_in_submotion(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test easymotion w command in submotion."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == text_yanked


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos",
    [
        ("", [Qt.Key_Space, Qt.Key_Space, "b"], 0),
        ("a b.c d;e\n", ["$", Qt.Key_Space, Qt.Key_Space, "b", "h"], 6),
        ("a b.c d;e\n", ["$", Qt.Key_Space, Qt.Key_Space, "b", "e", "h"], 7),
        ('a b.c d"e\n\nf', ["2j", Qt.Key_Space, Qt.Key_Space, "b", "y"], 2),
    ],
)
def test_easymotion_b_cmd_in_normal(vim_bot, text, cmd_list, cursor_pos):
    """Test easymotion b command in normal."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert editor.textCursor().position() == cursor_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        (
            "a b.c de",
            ["$", "y", Qt.Key_Space, Qt.Key_Space, "b", "l"],
            2,
            "a b.c de",
            '"',
            "b.c d",
        ),
        (
            "a b.c de\nf k",
            ["j", "$", "d", Qt.Key_Space, Qt.Key_Space, "b", "y"],
            2,
            "a k",
            '"',
            "b.c de\nf ",
        ),
        (
            "a b.c de\nf k",
            ["j", "$", "c", Qt.Key_Space, Qt.Key_Space, "b", "y"],
            2,
            "a k",
            '"',
            "b.c de\nf ",
        ),
    ],
)
def test_easymotion_b_cmd_in_submotion(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test easymotion b command in submotion."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == text_yanked


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos",
    [
        ("", [Qt.Key_Space, Qt.Key_Space, "j"], 0),
        ("a b.c d;", [Qt.Key_Space, Qt.Key_Space, "j", "l"], 1),
        ("a b.c d;\ne", [Qt.Key_Space, Qt.Key_Space, "j", "h"], 9),
        ("a b.c d;\n e", [Qt.Key_Space, Qt.Key_Space, "j", "h"], 10),
        ("a b.c d;\n e\n\nf", [Qt.Key_Space, Qt.Key_Space, "j", "l"], 13),
        (
            "a\nb\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nf",
            [Qt.Key_Space, Qt.Key_Space, "j", "h"],
            2,
        ),
    ],
)
def test_easymotion_j_cmd_in_normal(vim_bot, text, cmd_list, cursor_pos):
    """Test easymotion j command in normal."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert editor.textCursor().position() == cursor_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        (
            "a b\nkk dd\nd",
            ["y", Qt.Key_Space, Qt.Key_Space, "j", "h"],
            0,
            "a b\nkk dd\nd",
            '"',
            "a b\nkk dd\n",
        ),
        (
            "a b\nkk dd\nd",
            ["d", Qt.Key_Space, Qt.Key_Space, "j", "h"],
            0,
            "d",
            '"',
            "a b\nkk dd\n",
        ),
        (
            "a b\nkk dd\nd",
            ["c", Qt.Key_Space, Qt.Key_Space, "j", "h"],
            0,
            "\nd",
            '"',
            "a b\nkk dd\n",
        ),
    ],
)
def test_easymotion_j_cmd_in_submotion(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test easymotion j command in submotion."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == text_yanked


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos",
    [
        ("", [Qt.Key_Space, Qt.Key_Space, "k"], 0),
        ("a b.c d;", [Qt.Key_Space, Qt.Key_Space, "k", "l"], 1),
        (" a b.c d;\ne", ["j", Qt.Key_Space, Qt.Key_Space, "k", "h"], 1),
        ("a b.c d;\n e", ["j", Qt.Key_Space, Qt.Key_Space, "k", "h"], 0),
        ("a b.c d;\n e\n\nf", ["3j", Qt.Key_Space, Qt.Key_Space, "k", "l"], 0),
        (
            "a\nb\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nf",
            ["G", Qt.Key_Space, Qt.Key_Space, "k", "j"],
            31,
        ),
    ],
)
def test_easymotion_k_cmd_in_normal(vim_bot, text, cmd_list, cursor_pos):
    """Test easymotion k command in normal."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert editor.textCursor().position() == cursor_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        (
            "a b\nkk dd\nd",
            ["j", "y", Qt.Key_Space, Qt.Key_Space, "k", "h"],
            0,
            "a b\nkk dd\nd",
            '"',
            "a b\nkk dd\n",
        ),
        (
            "a b\nkk dd\nd",
            ["j", "d", Qt.Key_Space, Qt.Key_Space, "k", "h"],
            0,
            "d",
            '"',
            "a b\nkk dd\n",
        ),
        (
            "a b\nkk dd\nd",
            ["j", "c", Qt.Key_Space, Qt.Key_Space, "k", "h"],
            0,
            "\nd",
            '"',
            "a b\nkk dd\n",
        ),
    ],
)
def test_easymotion_k_cmd_in_submotion(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test easymotion k command in submotion."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == text_yanked


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos",
    [
        ("", [Qt.Key_Space, Qt.Key_Space, "f", "a"], 0),
        ("b a a\na", [Qt.Key_Space, Qt.Key_Space, "f", "a", "h"], 2),
        ("b a a\na", [Qt.Key_Space, Qt.Key_Space, "f", "a", "k"], 4),
    ],
)
def test_easymotion_f_cmd_in_normal(vim_bot, text, cmd_list, cursor_pos):
    """Test easymotion f command in normal."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert editor.textCursor().position() == cursor_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        (
            "import numpy as np",
            ["v", Qt.Key_Space, Qt.Key_Space, "f", "p", "h"],
            2,
            [0, 3],
        )
    ],
)
def test_easymotion_f_cmd_in_visual(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test easymotion w command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        (
            "a a.a aa",
            ["y", Qt.Key_Space, Qt.Key_Space, "f", "a", "l"],
            0,
            "a a.a aa",
            '"',
            "a a.a a",
        ),
        (
            "a a.a aa",
            ["d", Qt.Key_Space, Qt.Key_Space, "f", "a", "l"],
            0,
            "a",
            '"',
            "a a.a a",
        ),
        (
            "a a.a aa",
            ["c", Qt.Key_Space, Qt.Key_Space, "f", "a", "l"],
            0,
            "a",
            '"',
            "a a.a a",
        ),
    ],
)
def test_easymotion_f_cmd_in_submotion(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test easymotion f command in submotion."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == text_yanked


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos",
    [
        ("", [Qt.Key_Space, Qt.Key_Space, "F", "a"], 0),
        ("b a a\na", ["$", Qt.Key_Space, Qt.Key_Space, "F", "a", "h"], 2),
        ("b a a\na", ["j", Qt.Key_Space, Qt.Key_Space, "F", "b", "h"], 0),
    ],
)
def test_easymotion_F_cmd_in_normal(vim_bot, text, cmd_list, cursor_pos):
    """Test easymotion F command in normal."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert editor.textCursor().position() == cursor_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        (
            "import numpy as np",
            ["$", "v", Qt.Key_Space, Qt.Key_Space, "F", "p", "h"],
            10,
            [10, 18],
        )
    ],
)
def test_easymotion_F_cmd_in_visual(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test easymotion F command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        (
            "a a.a aa",
            ["$", "y", Qt.Key_Space, Qt.Key_Space, "F", "a", "l"],
            0,
            "a a.a aa",
            '"',
            "a a.a a",
        ),
        (
            "a a.a aa",
            ["$", "d", Qt.Key_Space, Qt.Key_Space, "F", "a", "l"],
            0,
            "a",
            '"',
            "a a.a a",
        ),
        (
            "a a.a aa",
            ["$", "c", Qt.Key_Space, Qt.Key_Space, "F", "a", "l"],
            0,
            "a",
            '"',
            "a a.a a",
        ),
    ],
)
def test_easymotion_F_cmd_in_submotion(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test easymotion F command in submotion."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == text_yanked



"""Expose Vim core classes.

This package gathers all classes implementing the Vim emulation logic.
Users should import them directly from ``spyder_okvim.vim``.

"""

from .cursor import VimCursor
from .label import InlineLabel
from .macro import MacroManager
from .search import SearchInfo
from .state import DotCmdInfo, FindInfo, InputCmdInfo, KeyInfo, RegisterInfo, VimState
from .status import VimStatus

__all__ = [
    "VimState",
    "FindInfo",
    "InputCmdInfo",
    "DotCmdInfo",
    "KeyInfo",
    "RegisterInfo",
    "SearchInfo",
    "MacroManager",
    "InlineLabel",
    "VimCursor",
    "VimStatus",
]





# -*- coding: utf-8 -*-
"""Mixin classes used by command executors."""

# Standard Libraries
from typing import Callable

# Third Party Libraries
from spyder.config.manager import CONF

# Project Libraries
from spyder_okvim.executor.executor_base import FUNC_INFO, RETURN_EXECUTOR_METHOD_INFO
from spyder_okvim.spyder.config import CONF_SECTION


class MovementMixin:
    """Provide cursor movement commands for various executor classes.

    Subclasses must define ``move_cursor`` and ``move_cursor_no_end`` attributes
    pointing to callables that move the cursor to the given position.  The
    ``move_cursor_no_end`` method is used for motions that should not place the
    cursor at the end of a line.
    """

    move_cursor: Callable[[int], None]
    move_cursor_no_end: Callable[[int], None]

    # ------------------------------------------------------------------
    # Generic commands shared by normal, visual and vline executors
    # ------------------------------------------------------------------
    def colon(self, num: int = 1, num_str: str = ""):
        """Enter ``:`` command-line mode."""
        self.vim_status.set_message("")
        return RETURN_EXECUTOR_METHOD_INFO(self.executor_colon, False)

    # ------------------------------------------------------------------
    # Basic cursor movement commands shared by multiple executors
    # ------------------------------------------------------------------
    def zero(self, num: int = 1, num_str: str = ""):
        """Move to the first column of the current line."""
        motion_info = self.helper_motion.zero()
        self.move_cursor(motion_info.cursor_pos)

    def h(self, num: int = 1, num_str: str = ""):
        """Move cursor left."""
        motion_info = self.helper_motion.h(num=num)
        self.move_cursor(motion_info.cursor_pos)

    def j(self, num: int = 1, num_str: str = ""):
        """Move cursor down."""
        motion_info = self.helper_motion.j(num=num)
        self.move_cursor(motion_info.cursor_pos)

    def k(self, num: int = 1, num_str: str = ""):
        """Move cursor up."""
        motion_info = self.helper_motion.k(num=num)
        self.move_cursor(motion_info.cursor_pos)

    def l(self, num: int = 1, num_str: str = ""):
        """Move cursor right."""
        motion_info = self.helper_motion.l(num=num)
        self.move_cursor_no_end(motion_info.cursor_pos)

    def H(self, num: int = 1, num_str: str = ""):
        """Move cursor to the top of the page."""
        motion_info = self.helper_motion.H(num=num)
        self.move_cursor(motion_info.cursor_pos)

    def M(self, num: int = 1, num_str: str = ""):
        """Move cursor to the middle of the page."""
        motion_info = self.helper_motion.M(num=num)
        self.move_cursor(motion_info.cursor_pos)

    def L(self, num: int = 1, num_str: str = ""):
        """Move cursor to the bottom of the page."""
        motion_info = self.helper_motion.L(num=num)
        self.move_cursor(motion_info.cursor_pos)

    def caret(self, num: int = 1, num_str: str = ""):
        """Move to the first non-blank character of the line."""
        motion_info = self.helper_motion.caret(num=num)
        self.move_cursor(motion_info.cursor_pos)

    def dollar(self, num: int = 1, num_str: str = ""):
        """Move cursor to the end of the line."""
        motion_info = self.helper_motion.dollar(num=num)
        self.move_cursor_no_end(motion_info.cursor_pos)

    def w(self, num: int = 1, num_str: str = ""):
        """Move to the next word."""
        motion_info = self.helper_motion.w(num=num)
        self.move_cursor_no_end(motion_info.cursor_pos)

    def W(self, num: int = 1, num_str: str = ""):
        """Move to the next WORD."""
        motion_info = self.helper_motion.W(num=num)
        self.move_cursor_no_end(motion_info.cursor_pos)

    def b(self, num: int = 1, num_str: str = ""):
        """Move to the previous word."""
        motion_info = self.helper_motion.b(num=num)
        self.move_cursor(motion_info.cursor_pos)

    def B(self, num: int = 1, num_str: str = ""):
        """Move to the previous WORD."""
        motion_info = self.helper_motion.B(num=num)
        self.move_cursor(motion_info.cursor_pos)

    def e(self, num: int = 1, num_str: str = ""):
        """Move to the end of the current word."""
        motion_info = self.helper_motion.e(num=num)
        self.move_cursor(motion_info.cursor_pos)

    def percent(self, num: int = 1, num_str: str = ""):
        """Go to the matching bracket."""
        motion_info = self.helper_motion.percent(num=num, num_str=num_str)
        self.move_cursor(motion_info.cursor_pos)

    def semicolon(self, num: int = 1, num_str: str = ""):
        """Repeat the last f/t search."""
        motion_info = self.helper_motion.semicolon(num=num, num_str=num_str)
        self.move_cursor(motion_info.cursor_pos)

    def comma(self, num: int = 1, num_str: str = ""):
        """Repeat the last f/t search in opposite direction."""
        motion_info = self.helper_motion.comma(num=num, num_str=num_str)
        self.move_cursor(motion_info.cursor_pos)

    def space(self, num: int = 1, num_str: str = ""):
        """Move cursor right (space key)."""
        motion_info = self.helper_motion.space(num=num)
        self.move_cursor_no_end(motion_info.cursor_pos)

    def backspace(self, num: int = 1, num_str: str = ""):
        """Move cursor left (backspace key)."""
        motion_info = self.helper_motion.backspace(num=num)
        self.move_cursor_no_end(motion_info.cursor_pos)

    def enter(self, num: int = 1, num_str: str = ""):
        """Move cursor down (enter key)."""
        motion_info = self.helper_motion.enter(num=num)
        self.move_cursor(motion_info.cursor_pos)


class SelectionMixin:
    """Common operations for visual and block-visual executors."""

    # Hooks provided by subclasses
    apply_motion_info_in_sel: Callable[[object], None]
    set_cursor_pos_in_sel: Callable[[int], None]
    paste_in_sel: Callable[[int], None]

    # ------------------------------------------------------------------
    # Selection manipulation helpers
    # ------------------------------------------------------------------
    def r(self, num: int = 1, num_str: str = ""):
        """Replace selection with a typed character."""
        executor_sub = self.executor_sub_r
        executor_sub.pos_start = self.get_pos_start_in_selection()
        executor_sub.pos_end = self.get_pos_end_in_selection()

        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            [
                FUNC_INFO(self.vim_status.to_normal, False),
                FUNC_INFO(lambda: self.set_cursor_pos(executor_sub.pos_start), False),
            ]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def greater(self, num: int = 1, num_str: str = ""):
        """Indent the selected text."""
        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()
        self.helper_action._indent(sel_start, sel_end)
        self.vim_status.to_normal()
        self.vim_status.cursor.draw_vim_cursor()

    def less(self, num: int = 1, num_str: str = ""):
        """Unindent the selected text."""
        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()
        self.helper_action._unindent(sel_start, sel_end)
        self.vim_status.to_normal()
        self.vim_status.cursor.draw_vim_cursor()

    def quote(self, num: int = 1, num_str: str = ""):
        """Switch register name."""
        executor_sub = self.executor_sub_register
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred([])
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def x(self, num: int = 1, num_str: str = ""):
        """Alias for ``d``."""
        self.d(num, num_str)

    def c(self, num: int = 1, num_str: str = ""):
        """Delete selection and enter insert mode."""
        sel_start, _ = self.helper_action.yank(None)
        self.helper_action.delete(None, is_insert=True)
        self.vim_status.to_normal()
        self.set_cursor_pos(sel_start)
        self.get_editor().setFocus()

    def s(self, num: int = 1, num_str: str = ""):
        """Replace selection using sneak if enabled."""
        use_sneak = CONF.get(CONF_SECTION, "use_sneak")
        if use_sneak:
            executor_sub = self.executor_sub_sneak
            self.set_parent_info_to_submode(executor_sub, num, num_str)
            executor_sub.set_func_list_deferred(
                [
                    FUNC_INFO(self.apply_motion_info_in_sel, True),
                    FUNC_INFO(
                        self.helper_motion.display_another_group_after_sneak, False
                    ),
                ]
            )
            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)
        else:
            self.c(num, num_str)

    def y(self, num: int = 1, num_str: str = ""):
        """Yank selection into register."""
        sel_start, _ = self.helper_action.yank(None, is_explicit=True)
        self.vim_status.to_normal()
        self.set_cursor_pos(sel_start)

    def slash(self, num: int = 1, num_str: str = ""):
        """Enter search submode and execute motion on return."""
        self.vim_status.set_message("")
        executor_sub = self.executor_sub_search
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_sel, True)]
        )
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, False)

    def n(self, num: int = 1, num_str: str = ""):
        """Jump to next search occurrence."""
        motion_info = self.helper_motion.n(num=num, num_str=num_str)
        self.set_cursor_pos_in_sel(motion_info.cursor_pos)

    def N(self, num: int = 1, num_str: str = ""):
        """Jump to previous search occurrence."""
        motion_info = self.helper_motion.N(num=num, num_str=num_str)
        self.set_cursor_pos_in_sel(motion_info.cursor_pos)

    def o(self, num: int = 1, num_str: str = ""):
        """Move cursor to the opposite end of the selection."""
        sel_start = self.vim_status.get_pos_start_in_selection()
        sel_end = self.vim_status.get_pos_end_in_selection()
        cur_pos = self.get_cursor().position()
        new_pos = (
            sel_end - 1
            if abs(sel_start - cur_pos) < abs(sel_end - cur_pos)
            else sel_start
        )
        self.set_cursor_pos(new_pos)

    def u(self, num: int = 1, num_str: str = ""):
        """Lowercase selected text."""
        self.helper_action.handle_case(None, "lower")

    def U(self, num: int = 1, num_str: str = ""):
        """Uppercase selected text."""
        self.helper_action.handle_case(None, "upper")

    def p(self, num: int = 1, num_str: str = ""):
        """Paste contents of register over the selection."""
        self.paste_in_sel(num)

    def P(self, num: int = 1, num_str: str = ""):
        """Paste contents of register over the selection."""
        self.paste_in_sel(num)

    def m(self, num: int = 1, num_str: str = ""):
        """Set a bookmark name."""
        executor_sub = self.executor_sub_alnum
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.vim_status.set_bookmark, True)]
        )
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def apostrophe(self, num: int = 1, num_str: str = ""):
        """Jump to a bookmark linewise and keep selection."""
        executor_sub = self.executor_sub_alnum
        self.set_parent_info_to_submode(executor_sub, num, num_str)

        def run(ch: str):
            if ch.isupper():
                cur = self.vim_status.get_editorstack().get_current_filename()
                self.vim_status.jump_to_bookmark(ch)
                new = self.vim_status.get_editorstack().get_current_filename()
                if cur == new:
                    cursor = self.get_cursor()
                    self.set_cursor_pos_in_sel(cursor.block().position())
                else:
                    cursor = self.get_cursor()
                    self.set_cursor_pos(cursor.block().position())
                    self.vim_status.to_normal()
            else:
                info = self.helper_motion.apostrophe(ch)
                if info.cursor_pos is not None:
                    self.set_cursor_pos_in_sel(info.cursor_pos)

        executor_sub.set_func_list_deferred([FUNC_INFO(run, True)])
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def backtick(self, num: int = 1, num_str: str = ""):
        """Jump to a bookmark charwise and keep selection."""
        executor_sub = self.executor_sub_alnum
        self.set_parent_info_to_submode(executor_sub, num, num_str)

        def run(ch: str):
            if ch.isupper():
                cur = self.vim_status.get_editorstack().get_current_filename()
                self.vim_status.jump_to_bookmark(ch)
                new = self.vim_status.get_editorstack().get_current_filename()
                if cur != new:
                    self.vim_status.to_normal()
            else:
                info = self.helper_motion.backtick(ch)
                if info.cursor_pos is not None:
                    self.set_cursor_pos_in_sel(info.cursor_pos)

        executor_sub.set_func_list_deferred([FUNC_INFO(run, True)])
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def run_easymotion(self, num: int = 1, num_str: str = ""):
        """Invoke the EasyMotion submode."""
        executor_sub = self.executor_sub_easymotion
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_sel, True)]
        )
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def asterisk(self, num: int = 1, num_str: str = ""):
        """Search forward for word under the cursor."""
        motion_info = self.helper_motion.asterisk(num)
        self.apply_motion_info_in_sel(motion_info)

    def sharp(self, num: int = 1, num_str: str = ""):
        """Search backward for word under the cursor."""
        motion_info = self.helper_motion.sharp(num)
        self.apply_motion_info_in_sel(motion_info)



"""Tests for the Marks."""

# Third Party Libraries
import pytest
from qtpy.QtCore import Qt

# Project Libraries
from spyder_okvim.vim import VimState


def test_bookmark_set_and_jump(vim_bot):
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    # set bookmark 'a'
    qtbot.keyClicks(cmd_line, "ma")
    # move cursor
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(2)
    assert editor.textCursor().position() == 2
    # jump to bookmark
    qtbot.keyClicks(cmd_line, "'a")
    assert editor.textCursor().position() == 0


def test_global_bookmark_and_jump(vim_bot):
    _, stack, editor0, vim, qtbot = vim_bot
    editor0.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    # set bookmark 'A' (persistent)
    qtbot.keyClicks(cmd_line, "mA")
    path0 = stack.get_current_filename()

    # switch to another file
    filenames = stack.get_filenames()
    other = next(p for p in filenames if p != path0)
    stack.set_current_filename(other)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(2)

    # jump to bookmark 'A' should switch file and position
    qtbot.keyClicks(cmd_line, "'A")
    assert stack.get_current_filename() == path0
    assert editor0.textCursor().position() == 0


def test_bookmark_visual_mode(vim_bot):
    """Set and jump to bookmark while in visual mode."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "vma")
    qtbot.keyPress(cmd_line, Qt.Key_Escape)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(2)
    qtbot.keyClicks(cmd_line, "'a")
    assert editor.textCursor().position() == 0


def test_bookmark_vline_mode(vim_bot):
    """Set and jump to bookmark while in vline mode."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "Vma")
    qtbot.keyPress(cmd_line, Qt.Key_Escape)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(2)
    qtbot.keyClicks(cmd_line, "'a")
    assert editor.textCursor().position() == 0


def test_local_bookmark_not_cross_file(vim_bot):
    """Local marks should not work in other files."""
    _, stack, editor0, vim, qtbot = vim_bot
    editor0.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    path0 = stack.get_current_filename()

    # switch to another file
    other = next(p for p in stack.get_filenames() if p != path0)
    stack.set_current_filename(other)
    editor_other = stack.get_current_editor()
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(2)
    pos_before = editor_other.textCursor().position()

    qtbot.keyClicks(cmd_line, "'a")
    assert editor_other.textCursor().position() == pos_before


def test_local_bookmark_not_cross_file_visual(vim_bot):
    """Local marks ignored in visual mode on other files."""
    _, stack, editor0, vim, qtbot = vim_bot
    editor0.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    path0 = stack.get_current_filename()

    other = next(p for p in stack.get_filenames() if p != path0)
    stack.set_current_filename(other)
    editor_other = stack.get_current_editor()
    qtbot.keyClicks(cmd_line, "v")
    pos_before = editor_other.textCursor().position()
    qtbot.keyClicks(cmd_line, "'a")
    assert editor_other.textCursor().position() == pos_before


def test_local_bookmark_not_cross_file_vline(vim_bot):
    """Local marks ignored in vline mode on other files."""
    _, stack, editor0, vim, qtbot = vim_bot
    editor0.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    path0 = stack.get_current_filename()

    other = next(p for p in stack.get_filenames() if p != path0)
    stack.set_current_filename(other)
    editor_other = stack.get_current_editor()
    qtbot.keyClicks(cmd_line, "V")
    pos_before = editor_other.textCursor().position()
    qtbot.keyClicks(cmd_line, "'a")
    assert editor_other.textCursor().position() == pos_before


def test_bookmark_removed_after_edit(vim_bot):
    """Jump should fail if mark line no longer exists."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(2)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    editor.set_text("a\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    qtbot.keyClicks(cmd_line, "'a")
    assert editor.textCursor().position() == 0


@pytest.mark.parametrize(
    "cmd_list,text_expected,cursor_pos,reg_expected",
    [
        ("y'a", "abcd\nefgh\nijkl\n", 0, "abcd\nefgh\nijkl\n"),
        ("y`a", "abcd\nefgh\nijkl\n", 2, "cd\nefgh\nijk"),
        ("d'a", "", 0, "abcd\nefgh\nijkl\n"),
        ("d`a", "abl\n", 2, "cd\nefgh\nijk"),
        ("c'a", "\n", 0, "abcd\nefgh\nijkl\n"),
        ("c`a", "abl\n", 2, "cd\nefgh\nijk"),
    ],
)
def test_mark_operations(vim_bot, cmd_list, text_expected, cursor_pos, reg_expected):
    """Test y/d/c commands with mark motions."""
    _, stack, editor, vim, qtbot = vim_bot

    # Ensure current file is the one associated with *editor*
    stack.set_current_filename(stack.get_filenames()[0])

    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "2lma")
    qtbot.keyClicks(cmd_line, "2j")
    qtbot.keyClicks(cmd_line, cmd_list)

    reg = vim.vim_cmd.vim_status.register_dict['"']
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == reg_expected


@pytest.mark.parametrize(
    "cmd_list,text_expected,cursor_pos,reg_expected",
    [
        ("y'a", "abcd\nefgh\nijkl\n", 0, "abcd\nefgh\nijkl\n"),
        ("y`a", "abcd\nefgh\nijkl\n", 0, "abcd\nefgh\nijk"),
        ("d'a", "", 0, "abcd\nefgh\nijkl\n"),
        ("d`a", "l\n", 0, "abcd\nefgh\nijk"),
        ("c'a", "\n", 0, "abcd\nefgh\nijkl\n"),
        ("c`a", "l\n", 0, "abcd\nefgh\nijk"),
    ],
)
def test_mark_operations_after(
    vim_bot, cmd_list, text_expected, cursor_pos, reg_expected
):
    """Operations when mark is after the cursor."""
    _, stack, editor, vim, qtbot = vim_bot

    stack.set_current_filename(stack.get_filenames()[0])
    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(12)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    qtbot.keyClicks(cmd_line, cmd_list)

    reg = vim.vim_cmd.vim_status.register_dict['"']
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == reg_expected


@pytest.mark.parametrize(
    "cmd_list,text_expected,cursor_pos,reg_expected",
    [
        ("y'A", "abcd\nefgh\nijkl\n", 0, "abcd\nefgh\nijkl\n"),
        ("y`A", "abcd\nefgh\nijkl\n", 2, "cd\nefgh\nijk"),
        ("d'A", "", 0, "abcd\nefgh\nijkl\n"),
        ("d`A", "abl\n", 2, "cd\nefgh\nijk"),
        ("c'A", "\n", 0, "abcd\nefgh\nijkl\n"),
        ("c`A", "abl\n", 2, "cd\nefgh\nijk"),
    ],
)
def test_global_mark_operations(
    vim_bot, cmd_list, text_expected, cursor_pos, reg_expected
):
    """Global mark operations when mark is before cursor."""
    _, stack, editor, vim, qtbot = vim_bot

    stack.set_current_filename(stack.get_filenames()[0])
    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "2lmA")
    qtbot.keyClicks(cmd_line, "2j")
    qtbot.keyClicks(cmd_line, cmd_list)

    reg = vim.vim_cmd.vim_status.register_dict['"']
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == reg_expected


@pytest.mark.parametrize(
    "cmd_list,text_expected,cursor_pos,reg_expected",
    [
        ("y'A", "abcd\nefgh\nijkl\n", 0, "abcd\nefgh\nijkl\n"),
        ("y`A", "abcd\nefgh\nijkl\n", 0, "abcd\nefgh\nijk"),
        ("d'A", "", 0, "abcd\nefgh\nijkl\n"),
        ("d`A", "l\n", 0, "abcd\nefgh\nijk"),
        ("c'A", "\n", 0, "abcd\nefgh\nijkl\n"),
        ("c`A", "l\n", 0, "abcd\nefgh\nijk"),
    ],
)
def test_global_mark_operations_after(
    vim_bot, cmd_list, text_expected, cursor_pos, reg_expected
):
    """Global mark operations when mark is after the cursor."""
    _, stack, editor, vim, qtbot = vim_bot

    stack.set_current_filename(stack.get_filenames()[0])
    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(12)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "mA")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    qtbot.keyClicks(cmd_line, cmd_list)

    reg = vim.vim_cmd.vim_status.register_dict['"']
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == reg_expected


@pytest.mark.parametrize("cmd", ["y'A", "y`A", "d'A", "d`A", "c'A", "c`A"])
def test_global_mark_operations_removed(vim_bot, cmd):
    """Global operations should do nothing if mark was removed."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("abcd\nefgh\nijkl\n")
    # Place mark on the last line so it disappears after editing
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(10)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "mA")
    editor.set_text("b")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    reg_before = vim.vim_cmd.vim_status.register_dict['"'].content
    qtbot.keyClicks(cmd_line, cmd)
    reg_after = vim.vim_cmd.vim_status.register_dict['"'].content
    assert editor.toPlainText() == "b"
    assert editor.textCursor().position() == 0
    assert reg_before == reg_after


@pytest.mark.parametrize("cmd", ["y'A", "y`A", "d'A", "d`A", "c'A", "c`A"])
def test_global_mark_operations_cross_file(vim_bot, cmd):
    """Global mark motions ignored in other files."""
    _, stack, editor0, vim, qtbot = vim_bot

    stack.set_current_filename(stack.get_filenames()[0])
    editor0.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "mA")
    other = next(p for p in stack.get_filenames() if p != stack.get_current_filename())
    stack.set_current_filename(other)
    editor_other = stack.get_current_editor()
    editor_other.set_text("x\ny\nz\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(2)
    reg_before = vim.vim_cmd.vim_status.register_dict['"'].content
    pos_before = editor_other.textCursor().position()
    qtbot.keyClicks(cmd_line, cmd)
    reg_after = vim.vim_cmd.vim_status.register_dict['"'].content
    assert editor_other.textCursor().position() == pos_before
    assert editor_other.toPlainText() == "x\ny\nz\n"
    assert reg_before == reg_after
    vim.vim_cmd.vim_status.reset_for_test()


@pytest.mark.parametrize("cmd", ["y'a", "y`a", "d'a", "d`a", "c'a", "c`a"])
def test_mark_operations_removed(vim_bot, cmd):
    """Operations should do nothing if mark was removed."""
    _, stack, editor, vim, qtbot = vim_bot
    stack.set_current_filename(stack.get_filenames()[0])
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(4)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    editor.set_text("b")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    reg_before = vim.vim_cmd.vim_status.register_dict['"'].content
    qtbot.keyClicks(cmd_line, cmd)
    reg_after = vim.vim_cmd.vim_status.register_dict['"'].content
    assert editor.toPlainText() == "b"
    assert editor.textCursor().position() == 0
    assert reg_before == reg_after


def test_global_bookmark_removed_after_edit(vim_bot):
    """Global mark should be cleared if its line is removed."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "mA")
    editor.set_text("b")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    qtbot.keyClicks(cmd_line, "'A")
    assert editor.textCursor().position() == 0
    assert "A" in vim.vim_cmd.vim_status.bookmarks_global


def test_global_bookmark_overwrite(vim_bot):
    """Setting mA in another file overwrites previous global mark."""
    _, stack, editor0, vim, qtbot = vim_bot
    stack.set_current_filename(stack.get_filenames()[0])
    editor0 = stack.get_current_editor()
    editor0.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "mA")
    other = next(p for p in stack.get_filenames() if p != stack.get_current_filename())
    stack.set_current_filename(other)
    editor_other = stack.get_current_editor()
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(2)
    qtbot.keyClicks(cmd_line, "mA")
    stack.set_current_filename(stack.get_filenames()[0])
    qtbot.keyClicks(cmd_line, "'A")
    assert stack.get_current_filename() == other
    stack.set_current_filename(stack.get_filenames()[0])


@pytest.mark.parametrize("cmd,pos_expected", [("'a", 0), ("`a", 2)])
def test_jump_mark_visual_before_after(vim_bot, cmd, pos_expected):
    """Jump to mark from visual mode with mark before and after cursor."""
    _, stack, editor, vim, qtbot = vim_bot
    stack.set_current_filename(stack.get_filenames()[0])
    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "2lma")
    qtbot.keyClicks(cmd_line, "2j")
    qtbot.keyClicks(cmd_line, "v")
    qtbot.keyClicks(cmd_line, cmd)
    assert vim.vim_cmd.vim_status.vim_state == VimState.VISUAL
    assert editor.textCursor().position() == pos_expected


@pytest.mark.parametrize("cmd", ["'a", "`a"])
def test_jump_mark_visual_removed(vim_bot, cmd):
    """Jump should do nothing if mark line was removed."""
    _, stack, editor, vim, qtbot = vim_bot
    stack.set_current_filename(stack.get_filenames()[0])
    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(10)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    editor.set_text("x")
    qtbot.keyClicks(cmd_line, "v")
    qtbot.keyClicks(cmd_line, cmd)
    assert editor.textCursor().position() == 0
    assert vim.vim_cmd.vim_status.vim_state == VimState.VISUAL
    assert vim.vim_cmd.vim_status.vim_state == VimState.VISUAL


@pytest.mark.parametrize("cmd,pos_expected", [("'A", 0), ("`A", 2)])
def test_jump_global_mark_visual(vim_bot, cmd, pos_expected):
    """Jump to global mark from visual mode."""
    _, stack, editor, vim, qtbot = vim_bot
    stack.set_current_filename(stack.get_filenames()[0])
    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "2lmA")
    qtbot.keyClicks(cmd_line, "2j")
    qtbot.keyClicks(cmd_line, "v")
    qtbot.keyClicks(cmd_line, cmd)
    assert stack.get_current_filename() == stack.get_filenames()[0]
    assert editor.textCursor().position() == pos_expected
    assert vim.vim_cmd.vim_status.vim_state == VimState.VISUAL


@pytest.mark.parametrize("cmd,pos_expected", [("'A", 0), ("`A", 2)])
def test_jump_global_mark_visual_cross_file(vim_bot, cmd, pos_expected):
    """Global mark jump from another file in visual mode."""
    _, stack, editor0, vim, qtbot = vim_bot
    stack.set_current_filename(stack.get_filenames()[0])
    editor0.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "2lmA")
    other = next(p for p in stack.get_filenames() if p != stack.get_current_filename())
    stack.set_current_filename(other)
    editor_other = stack.get_current_editor()
    qtbot.keyClicks(cmd_line, "v")
    qtbot.keyClicks(cmd_line, cmd)
    assert stack.get_current_filename() == stack.get_filenames()[0]
    assert editor0.textCursor().position() == pos_expected
    assert vim.vim_cmd.vim_status.vim_state == VimState.NORMAL


@pytest.mark.parametrize("cmd", ["'A", "`A"])
def test_jump_global_mark_visual_removed(vim_bot, cmd):
    """Global mark jump does nothing if its line was removed."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("abcd\nefgh\nijkl\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(10)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "mA")
    editor.set_text("x")
    qtbot.keyClicks(cmd_line, "v")
    qtbot.keyClicks(cmd_line, cmd)
    assert editor.textCursor().position() == 0



# -*- coding: utf-8 -*-
"""Macro recording and playback utilities."""

from collections import defaultdict

from qtpy.QtCore import QObject
from qtpy.QtGui import QKeyEvent

from .state import KeyInfo


class MacroManager:
    """Store and play back recorded macros."""

    def __init__(self):
        #: Mapping of register names to recorded :class:`KeyInfo` instances
        self.registers = defaultdict(list)
        #: Whether a macro is currently being recorded
        self.is_recording = False
        #: Register being used for recording
        self.reg_name_for_record = ""
        #: Editor currently connected to receive key events during recording
        self.editor_connected = None
        #: Register to execute
        self.reg_name_for_execute = ""
        #: Number of times to execute the macro
        self.num_execute = 0

    def set_info_for_execute(self, register, count):
        """Configure macro execution.

        Args:
            register: Name of the register to execute.
            count: Number of times to repeat execution.
        """
        self.reg_name_for_execute = register
        self.num_execute = count

    def start_record(self, register):
        """Begin recording keystrokes.

        Args:
            register: Name of the register to store the recording in.
        """
        self.reg_name_for_record = register
        self.is_recording = True

        if register.lower() == register:
            self.registers[register] = list()

    def stop_record(self):
        """Finish recording and clean up."""
        # Remove the trailing ``q`` when finishing recording
        self.remove_last_key("q")
        self.reg_name_for_record = ""
        self.is_recording = False

    def remove_last_key(self, text):
        """Remove a trailing key from the recording if it matches ``text``.

        Args:
            text: Character to remove from the end of the recorded sequence.
        """
        key_list = self.registers[self.reg_name_for_record]
        if key_list:
            last_key = key_list[-1]
            if last_key.text == text:
                self.registers[self.reg_name_for_record] = key_list[:-1]

    def add_vim_keyevent(self, event: QKeyEvent):
        """Record a key event coming from the Vim command line."""
        if self.is_recording:
            self.registers[self.reg_name_for_record].append(
                KeyInfo(event.key(), event.text(), event.modifiers(), 0)
            )

    def add_editor_keyevent(self, event: QKeyEvent):
        """Record a key event originating in the editor."""
        self.registers[self.reg_name_for_record].append(
            KeyInfo(event.key(), event.text(), event.modifiers(), 1)
        )

    def connect_to_editor(self, editor: QObject, slot):
        """Start receiving key events from ``editor``.

        Args:
            editor: Editor emitting ``sig_key_pressed``.
            slot: Slot connected to the editor signal.
        """
        self.editor_connected = editor
        editor.sig_key_pressed.connect(slot)

    def disconnect_from_editor(self, slot):
        """Stop receiving key events from the previously connected editor."""
        editor = self.editor_connected
        if editor:
            try:
                editor.sig_key_pressed.disconnect(slot)
            except (TypeError, RuntimeError):
                # Already disconnected or editor deleted
                pass
        self.editor_connected = None







# -*- coding: utf-8 -*-
"""Tests for the executor_visual."""

# Third Party Libraries
import pytest
from qtpy.QtCore import Qt
from spyder.config.manager import CONF

# Project Libraries
from spyder_okvim.spyder.config import CONF_SECTION
from spyder_okvim.vim import VimState


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("import numpy as np", ["v"], 0, [0, 1]),
        ("import numpy as np", ["v", "5l"], 5, [0, 6]),
        ("import numpy as np", ["5l", "v"], 5, [5, 6]),
        ("import numpy as np", ["5l", "v", "3l"], 8, [5, 9]),
        ("import numpy as np", ["5l", "v", "3l", "7h"], 1, [1, 6]),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["v", "j"],
            19,
            [0, 20],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["v", "j", "5l"],
            24,
            [0, 25],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["v", "j", "5l", "3h"],
            21,
            [0, 22],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["l", "v", "2j"],
            52,
            [1, 53],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["l", "v", "2j", "4l"],
            56,
            [1, 57],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["l", "v", "2j", "4l", "k"],
            24,
            [1, 25],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["j", "5l", "v", "k"],
            5,
            [5, 25],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["j", "5l", "v", "k", "5h"],
            0,
            [0, 25],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["j", "5l", "v", "k", "5h", "j"],
            19,
            [19, 25],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["j", "5l", "v", "k", "5h", "2j"],
            51,
            [24, 52],
        ),
        (
            """
import matplotlib.pyplot as plt
import scipy.scipy as sc
""",
            ["v", "3j", "30l"],
            58,
            [0, 58],
        ),
    ],
)
def test_v_cmd(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test v command."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("import numpy as np", ["v", "0"], 0, [0, 1]),
        ("import numpy as np", ["5l", "v", "0"], 0, [0, 6]),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["v", "j", "0"],
            19,
            [0, 20],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["v", "j", "5l", "0"],
            19,
            [0, 20],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["l", "v", "2j", "0"],
            51,
            [1, 52],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["l", "v", "2j", "4l", "k", "0"],
            19,
            [1, 20],
        ),
        (
            """
import matplotlib.pyplot as plt
import scipy.scipy as sc
""",
            ["v", "3j", "0"],
            58,
            [0, 58],
        ),
    ],
)
def test_zero_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test 0 command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("   import numpy as np", ["v", "^"], 3, [0, 4]),
        ("   import numpy as np", ["10l", "v", "^"], 3, [3, 11]),
    ],
)
def test_caret_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test ^ command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("import numpy as np", ["v", "$"], 18, [0, 18]),
    ],
)
def test_dollar_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test $ command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34", ["v", "w"], 3, [0, 4]),
        ("01 34", ["v", "w", "o"], 0, [0, 4]),
        ("01 34", ["v", "w", "o", "o"], 3, [0, 4]),
        ("\n", ["j", "v"], 1, [1, 1]),
        ("\n", ["j", "v", "o"], 1, [1, 1]),
        ("01 34\n6", ["v", "j"], 6, [0, 7]),
        ("01 34\n6", ["v", "j", "o"], 0, [0, 7]),
    ],
)
def test_o_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test o command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("", ["v", "J"], "", 0),
        ("\n\n", ["3j", "v", "J"], "\n\n", 2),
        ("0\n23", ["j", "l", "v", "k", "J"], "0 23", 1),
        ("0\n2\n4\n6\n8\n", ["v", "2j", "J"], "0 2 4\n6\n8\n", 3),
        ("0\n2\n4\n6\n8\n", ["v", "2j", "J", "."], "0 2 4 6 8\n", 7),
    ],
)
def test_J_cmd_in_v(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test J command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34", ["v", "w"], 3, [0, 4]),
        ("01 34", ["v", "2w"], 5, [0, 5]),
        ("01 34\na\n", ["v", "3w"], 8, [0, 8]),
        ("01 34\n  a\n", ["v", "3w"], 10, [0, 10]),
    ],
)
def test_w_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test w command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["v", "W"], 0, [0, 0]),
        ("029.d98@jl 34", ["v", "W"], 11, [0, 12]),
        ("029.d98@jl 34", ["v", "2W"], 13, [0, 13]),
        ("029.d98@jl 34\na", ["v", "2W"], 14, [0, 15]),
        ("029.d98@jl 34\n  a", ["v", "2W"], 16, [0, 17]),
    ],
)
def test_W_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test W command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34", ["$", "v", "b"], 3, [3, 5]),
    ],
)
def test_b_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test b command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01.34", ["$", "v", "B"], 0, [0, 5]),
    ],
)
def test_B_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test B command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34", ["v", "e"], 1, [0, 2]),
    ],
)
def test_e_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test e command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("0\n2\n4\n", ["v", "2G"], 2, [0, 3]),
        ("0\n     \n8\n", ["v", "2G"], 6, [0, 7]),
        ("0\n2\n4\n", ["v", "G"], 6, [0, 6]),
        ("0\n2\n4\n     a", ["v", "G"], 11, [0, 12]),
    ],
)
def test_G_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test G command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("0\n2\n4\n", ["v", "2gg"], 2, [0, 3]),
        ("0\n     \n8\n", ["v", "2gg"], 6, [0, 7]),
        ("    0\n2\n4\n", ["4j", "v", "gg"], 4, [4, 10]),
    ],
)
def test_gg_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test gg command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos",
    [
        ("0\n2\n4\n", ["v", ":", "2", Qt.Key_Return], 4),
        ("0\n     \n8\n", ["v", ":", "2", Qt.Key_Return], 8),
        ("0\n\n2\n4\n", ["v", ":", "3", Qt.Key_Return], 5),
        (
            "aaa\nbbb\nccc\nddd\neee\n",
            ["j", "v", ":", "3", Qt.Key_Return],
            16,
        ),
        (
            "aaa\nbbb\nccc\nddd\neee\n",
            ["j", "l", "v", ":", "3", Qt.Key_Return],
            17,
        ),
    ],
)
def test_colon_num_cmd_in_v(vim_bot, text, cmd_list, cursor_pos):
    """Test :num command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert not editor.get_extra_selections("vim_selection")


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("abcde", ["v", "~"], "Abcde", 0),
        ("abcde\na", ["l", "v", "$", "~"], "aBCDE\na", 1),
        ("abcde\na", ["l", "v", "3l", "~", "0", "."], "AbcdE\na", 0),
        ("", ["v", "~"], "", 0),
    ],
)
def test_tilde_cmd_in_v(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test ~ command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["v", "%"], 0, [0, 0]),
        ("\n", ["j", "v", "%"], 1, [1, 1]),
        (" ()", ["v", "%"], 2, [0, 3]),
        (" ()", ["v", "%", "%"], 1, [0, 2]),
    ],
)
def test_percent_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test % command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["v", "f", "r"], 0, [0, 0]),
        ("\n", ["j", "v", "f", "r"], 1, [1, 1]),
        (" rr", ["v", "f", "r"], 1, [0, 2]),
        (" rr", ["v", "f", "r", ";"], 2, [0, 3]),
        (" rr", ["v", "f", "r", ";", ","], 1, [0, 2]),
    ],
)
def test_f_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test f command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["v", "F", "r"], 0, [0, 0]),
        ("\n", ["j", "v", "F", "r"], 1, [1, 1]),
        (" rr ", ["v", "$", "F", "r"], 2, [0, 3]),
        (" rr ", ["v", "$", "F", "r", ";"], 1, [0, 2]),
        (" rr ", ["v", "$", "F", "r", ";", ","], 2, [0, 3]),
    ],
)
def test_F_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test F command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["v", "s", "aa"], 0, [0, 0]),
        ("", ["v", ";"], 0, [0, 0]),
        ("", ["v", ","], 0, [0, 0]),
        ("\n", ["v", "j", "s", "rr"], 1, [0, 1]),
        ("d\ndhr Dhr dhr", ["v", "sdh"], 2, [0, 3]),
        ("d\ndhr Dhr dhr", ["v", "sdh", ";"], 10, [0, 11]),
        ("d\ndhr Dhr dhr", ["v", "sdh", ";,"], 2, [0, 3]),
    ],
)
def test_sneak_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test sneak command in visual."""
    CONF.set(CONF_SECTION, "use_sneak", True)
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["v", "t", "r"], 0, [0, 0]),
        ("\n", ["j", "v", "t", "r"], 1, [1, 1]),
        ("  rr", ["v", "t", "r"], 1, [0, 2]),
        ("  rr", ["v", "t", "r", ";"], 2, [0, 3]),
        ("  rrrr", ["v", "t", "r", "4;"], 4, [0, 5]),
        ("  rrrr", ["v", "t", "r", "4;", ","], 3, [0, 4]),
    ],
)
def test_t_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test t command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["v", "T", "r"], 0, [0, 0]),
        ("r\n", ["j", "v", "T", "r"], 2, [2, 2]),
        ("  rr  ", ["v", "$", "T", "r"], 4, [0, 5]),
        ("  rr  ", ["v", "$", "T", "r", ";"], 3, [0, 4]),
        ("  rrrr", ["v", "$", "T", "r", "4;"], 3, [0, 4]),
        ("  rrrr", ["v", "$", "T", "r", "4;", ","], 4, [0, 5]),
    ],
)
def test_T_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test T command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("", ["v", "r", "r"], "", 0),
        ("1\n", ["j", "v", "r", "r"], "1\n", 2),
        ("\n\na", ["j", "v", "r", "r"], "\n\na", 1),
        ("a", ["v", "r", "r"], "r", 0),
        (" a\nbc\n", ["l", "v", "j", "r", "r"], " r\nrr\n", 1),
        (" a\nbc\nde", ["l", "v", "j", "r", "r"], " r\nrr\nde", 1),
        (" a\nbc\nde", ["l", "v", "j", "r", "r", "j", "."], " r\nrr\nrr", 4),
    ],
)
def test_r_cmd_in_v(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test r command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")

    assert sel == []
    assert cmd_line.text() == ""
    assert vim.vim_cmd.vim_status.vim_state == VimState.NORMAL
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("ABCDE", ["v", "u"], "aBCDE", 0),
        ("ABCDE\nA", ["l", "v", "$", "u"], "Abcde\nA", 1),
        ("ABCDE\nA", ["l", "v", "3l", "u", "0", "."], "abcde\nA", 0),
        ("", ["v", "u"], "", 0),
    ],
)
def test_u_cmd_in_v(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test u command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("ABCDE", ["v", "g", "u"], "aBCDE", 0),
        ("ABCDE\nA", ["l", "v", "$", "g", "u"], "Abcde\nA", 1),
        ("ABCDE\nA", ["l", "v", "3l", "g", "u", "0", "."], "abcde\nA", 0),
        ("", ["v", "g", "u"], "", 0),
    ],
)
def test_gu_cmd_in_v(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test gu command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("abcde", ["v", "U"], "Abcde", 0),
        ("abcde\na", ["l", "v", "$", "U"], "aBCDE\na", 1),
        ("abcde\na", ["l", "v", "3l", "U", "0", "."], "ABCDE\na", 0),
        ("", ["v", "U"], "", 0),
    ],
)
def test_U_cmd_in_v(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test U command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("abcde", ["v", "g", "U"], "Abcde", 0),
        ("abcde\na", ["l", "v", "$", "g", "U"], "aBCDE\na", 1),
        ("abcde\na", ["l", "v", "3l", "g", "U", "0", "."], "ABCDE\na", 0),
        ("", ["v", "g", "U"], "", 0),
    ],
)
def test_gU_cmd_in_v(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test gU command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("abCde", ["v", "g", "~"], "AbCde", 0),
        ("abCde\na", ["l", "v", "$", "g", "~"], "aBcDE\na", 1),
        ("abCde\na", ["l", "v", "3l", "g", "~", "0", "."], "AbCdE\na", 0),
        ("", ["v", "g", "~"], "", 0),
    ],
)
def test_gtilde_cmd_in_v(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test g~ command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("", ["v", ">"], "", 0),
        ("abcde", ["2l", "v", ">"], "    abcde", 4),
        (" abcde\na", ["v", ">"], "     abcde\na", 5),
        ("a\n\na", ["v", "2j", ">"], "    a\n\n    a", 4),
    ],
)
def test_greater_cmd_in_visual(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test > command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("", ["v", "<"], "", 0),
        ("    abcde", ["2l", "v", "<"], "abcde", 0),
        ("     abcde\na", ["v", "<"], " abcde\na", 1),
        ("    a\n\n    a", ["v", "2j", "<"], "a\n\na", 0),
    ],
)
def test_less_cmd_in_visual(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test < command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ('"dkwh"  "sdalkj"', ["8l", "v", "i", '"'], 14, [9, 15]),
        ('"dkwh"  "sdalkj"', ["7l", "v", "i", '"'], 7, [6, 8]),
        ('"dkwh"sd"', ["$", "v", "i", '"'], 8, [8, 9]),
        ("", ["v", "i", '"'], 0, [0, 0]),
        ('"0"', ["v", "i", '"'], 1, [1, 2]),
        ('"0"', ["l", "v", "i", '"'], 1, [1, 2]),
        ('"0"', ["2l", "v", "i", '"'], 1, [1, 2]),
        ('"0" ', ["3l", "v", "i", '"'], 3, [3, 4]),
        (' "0" ', ["v", "i", '"'], 2, [2, 3]),
        ("'0'", ["v", "i", "'"], 1, [1, 2]),
        ("'0'", ["l", "v", "i", "'"], 1, [1, 2]),
        ("'0'", ["2l", "v", "i", "'"], 1, [1, 2]),
        ("'0' ", ["3l", "v", "i", "'"], 3, [3, 4]),
        (" '0' ", ["v", "i", "'"], 2, [2, 3]),
    ],
)
def test_iquote_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test i" i' command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ('"dkwh"  "sdalkj"', ["8l", "v", "a", '"'], 15, [6, 16]),
        ('"dkwh"  "sdalkj"', ["7l", "v", "a", '"'], 8, [5, 9]),
        ('"dkwh"sd"', ["$", "v", "a", '"'], 8, [8, 9]),
        ("", ["v", "a", '"'], 0, [0, 0]),
        ('"0"', ["v", "a", '"'], 2, [0, 3]),
        ('"0"a', ["l", "v", "a", '"'], 2, [0, 3]),
        ('"0"   ', ["2l", "v", "a", '"'], 5, [0, 6]),
        ('"0"   a', ["2l", "v", "a", '"'], 5, [0, 6]),
        ('"0"   \na', ["2l", "v", "a", '"'], 5, [0, 6]),
        ('"0" ', ["3l", "v", "a", '"'], 3, [3, 4]),
        (' "0" ', ["v", "a", '"'], 4, [1, 5]),
        ("'0'", ["v", "a", "'"], 2, [0, 3]),
        ("'0'", ["l", "v", "a", "'"], 2, [0, 3]),
        ("'0'", ["2l", "v", "a", "'"], 2, [0, 3]),
        ("'0' ", ["3l", "v", "a", "'"], 3, [3, 4]),
    ],
)
def test_aquote_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test a" a' command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("kk 2 3", ["v", "i", "w"], 1, [0, 2]),
        ("kk 2 3", ["l", "v", "i", "w"], 1, [0, 2]),
        ("kk 2 3", ["v", "2i", "w"], 2, [0, 3]),
        ("kk 2 3", ["l", "v", "5i", "w"], 5, [0, 6]),
        ("kk   2 3", ["3l", "v", "i", "w"], 4, [2, 5]),
        ("kk   2 3", ["2l", "v", "i", "w"], 4, [2, 5]),
        ("kk   2 3", ["4l", "v", "i", "w"], 4, [2, 5]),
        ("kk   2 3", ["2l", "v", "2i", "w"], 5, [2, 6]),
        ("kk   2 3\na", ["2l", "v", "4i", "w"], 7, [2, 8]),
        ("kk   2 3\na", ["2l", "v", "5i", "w"], 9, [2, 10]),
        ("kk   2 3\n   a", ["2l", "v", "5i", "w"], 11, [2, 12]),
        ("kk   2 3\na", ["v", "6i", "w"], 9, [0, 10]),
        ("abc.def", ["v", "i", "w"], 2, [0, 3]),
        ("abc.def", ["v", "2i", "w"], 3, [0, 4]),
        ("abc.def", ["v", "3i", "w"], 6, [0, 7]),
        ("abc.def", ["3l", "v", "i", "w"], 3, [3, 4]),
    ],
)
def test_iw_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test iw command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["v", "i", "W"], 0, [0, 0]),
        ("\n\n", ["j", "v", "i", "W"], 1, [1, 2]),
        ("fig.add_subplot(111)  # ", ["4l", "v", "i", "W"], 19, [0, 20]),
        ("fig.add_subplot(111)  # ", ["4l", "v", "2i", "W"], 21, [0, 22]),
        ("fig.add_subplot(111)  # ", ["4l", "v", "3i", "W"], 22, [0, 23]),
        ("fig.add_subplot(111)  # ", ["20l", "v", "i", "W"], 21, [20, 22]),
        ("fig.add_subplot(111)  # ", ["20l", "v", "2i", "W"], 22, [20, 23]),
        (" # \n 1", ["l", "v", "3i", "W"], 4, [1, 5]),
    ],
)
def test_iW_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test iW command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("kk 2 3", ["v", "a", "w"], 2, [0, 3]),
        ("kk 2 3", ["l", "v", "a", "w"], 2, [0, 3]),
        ("kk 2 3", ["v", "2a", "w"], 4, [0, 5]),
        ("kk 2 3", ["v", "3a", "w"], 5, [0, 6]),
        ("kk 2 3 ", ["v", "3a", "w"], 6, [0, 7]),
        ("kk 2 3\n", ["v", "4a", "w"], 6, [0, 7]),
        ("kk 2 3 \n", ["v", "4a", "w"], 7, [0, 8]),
        ("kk 2 3 \na", ["v", "4a", "w"], 8, [0, 9]),
        ("kk 2 3\na ", ["v", "4a", "w"], 8, [0, 9]),
        ("kk   2 3", ["3l", "v", "a", "w"], 5, [2, 6]),
        ("kk   2 3", ["2l", "v", "a", "w"], 5, [2, 6]),
        ("kk   2 3", ["4l", "v", "a", "w"], 5, [2, 6]),
        ("kk   2 3 ", ["2l", "v", "2a", "w"], 7, [2, 8]),
        ("kk   2 3 4 ", ["2l", "v", "3a", "w"], 9, [2, 10]),
        ("kk   2 3\na", ["2l", "v", "3a", "w"], 9, [2, 10]),
        ("kk   2 3\n a", ["2l", "v", "3a", "w"], 10, [2, 11]),
        ("kk   2 3 \n a", ["2l", "v", "3a", "w"], 11, [2, 12]),
        ("kk   2 3 \n a", ["5l", "v", "a", "w"], 6, [5, 7]),
        ("kk   2 3 \n a", ["5l", "v", "2a", "w"], 8, [5, 9]),
        ("kk   2 3 \n a", ["5l", "v", "3a", "w"], 10, [5, 11]),
    ],
)
def test_aw_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test aw command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["v", "a", ")"], 0, [0, 0]),
        (" () ", ["v", "a", ")"], 2, [1, 3]),
        (" () ", ["l", "v", "a", ")"], 2, [1, 3]),
        (" (01234) ", ["3l", "v", "a", ")"], 7, [1, 8]),
        (" (01234) ", ["3l", "v", "a", "("], 7, [1, 8]),
        (" (01234) ", ["3l", "v", "a", "b"], 7, [1, 8]),
        (" [01234] ", ["3l", "v", "a", "["], 7, [1, 8]),
        (" [01234] ", ["3l", "v", "a", "]"], 7, [1, 8]),
        (" {01234} ", ["3l", "v", "a", "{"], 7, [1, 8]),
        (" {01234} ", ["3l", "v", "a", "B"], 7, [1, 8]),
        (" {01234} ", ["3l", "v", "a", "}"], 7, [1, 8]),
        (" {01234} ", ["l", "v", "a", "}"], 7, [1, 8]),
        (" {01234} ", ["7l", "v", "a", "}"], 7, [1, 8]),
        (" {\n(\nasdf)\nasdf} ", ["3l", "v", "a", "}"], 15, [1, 16]),
        (" (((kk)))", ["l", "v", "a", "("], 8, [1, 9]),
        (" (((kk)))", ["$", "v", "a", "("], 8, [1, 9]),
    ],
)
def test_a_bracket_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test a_bracket command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["v", "i", ")"], 0, [0, 0]),
        (" () ", ["v", "i", ")"], 0, [0, 1]),
        (" () ", ["l", "v", "i", ")"], 1, [1, 2]),
        (" (01234) ", ["3l", "v", "i", ")"], 6, [2, 7]),
        (" (01234) ", ["3l", "v", "i", "("], 6, [2, 7]),
        (" (01234) ", ["3l", "v", "i", "b"], 6, [2, 7]),
        (" [01234] ", ["3l", "v", "i", "["], 6, [2, 7]),
        (" [01234] ", ["3l", "v", "i", "]"], 6, [2, 7]),
        (" {01234} ", ["3l", "v", "i", "{"], 6, [2, 7]),
        (" {01234} ", ["3l", "v", "i", "}"], 6, [2, 7]),
        (" {01234} ", ["3l", "v", "i", "B"], 6, [2, 7]),
        (" {01234} ", ["l", "v", "i", "}"], 6, [2, 7]),
        (" {01234} ", ["7l", "v", "i", "}"], 6, [2, 7]),
        (" {\n(\nasdf)\nasdf} ", ["3l", "v", "i", "}"], 14, [2, 15]),
    ],
)
def test_i_bracket_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test i_bracket command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, register_name, text_yanked",
    [
        ("a", ["v", "y"], 0, '"', "a"),
        ("abcd", ["l", "v", "2l", "y"], 1, '"', "bcd"),
        ("abcd", ["l", "v", "2l", '"', "0", "y"], 1, "0", "bcd"),
        ("abcd\ne", ["l", "v", "2l", "j", '"', "a", "y"], 1, "a", "bcd\ne"),
    ],
)
def test_y_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, register_name, text_yanked):
    """Test y command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[register_name]
    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == text_yanked
    assert reg.type == VimState.NORMAL
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None
    if register_name == '"':
        reg0 = vim.vim_cmd.vim_status.register_dict["0"]
        assert reg0.content == text_yanked


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected",
    [
        ("ak", ["v", "p"], 0, "k"),
        ("ak", ["v", "l", "y", "v", "p"], 1, "akk"),
        ("ak", ["v", "l", "y", "v", "P"], 1, "akk"),
        ("ak", ["v", "l", "y", "v", "2p"], 3, "akakk"),
        ("ab\ncd\nef", ["v", "j", "l", "y", "j", "v", "p"], 3, "ab\nab\ncdd\nef"),
        ("ab\ncd\nef", ["V", "y", "j", "v", "p"], 4, "ab\n\nab\nd\nef"),
    ],
)
def test_p_P_cmd_in_visual(vim_bot, text, cmd_list, cursor_pos, text_expected):
    """Test p command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        ("ak", ["v", "d"], 0, "k", '"', "a"),
        ("ak", ["v", "l", "d"], 0, "", '"', "ak"),
        ("ab\nc\nde", ["v", "2j", "l", "d"], 0, "", '"', "ab\nc\nde"),
        ("ab\nc\nde", ["l", "v", "d"], 0, "a\nc\nde", '"', "b"),
    ],
)
def test_d_cmd_in_visual(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test d command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected
    assert reg.content == text_yanked
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        ("ak", ["v", "x"], 0, "k", '"', "a"),
        ("ak", ["v", "l", "x"], 0, "", '"', "ak"),
        ("ab\nc\nde", ["v", "2j", "l", "x"], 0, "", '"', "ab\nc\nde"),
        ("ab\nc\nde", ["l", "v", "x"], 0, "a\nc\nde", '"', "b"),
    ],
)
def test_x_cmd_in_visual(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test x command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected
    assert reg.content == text_yanked
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        ("ak", ["v", "c"], 0, "k", '"', "a"),
        ("ak", ["v", "l", "c"], 0, "", '"', "ak"),
        ("ab\nc\nde", ["v", "2j", "l", "c"], 0, "", '"', "ab\nc\nde"),
        ("ab\nc\nde", ["l", "v", "c"], 1, "a\nc\nde", '"', "b"),
    ],
)
def test_c_cmd_in_visual(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test c command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected
    assert reg.content == text_yanked
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        ("ak", ["v", "s"], 0, "k", '"', "a"),
        ("ak", ["v", "l", "s"], 0, "", '"', "ak"),
        ("ab\nc\nde", ["v", "2j", "l", "s"], 0, "", '"', "ab\nc\nde"),
        ("ab\nc\nde", ["l", "v", "s"], 1, "a\nc\nde", '"', "b"),
    ],
)
def test_s_cmd_in_visual(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test s command in visual."""
    CONF.set(CONF_SECTION, "use_sneak", False)
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected
    assert reg.content == text_yanked
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("a", ["v", "/", "b", "\r"], 0, [0, 1]),
        ("a", ["v", "/", "b", "\r", "n"], 0, [0, 1]),
        (
            " dhr\n  dhrwodn\n\ndhrwodn\n   dhrwodn",
            ["v", "/", "d", "h", "r", Qt.Key_Return],
            1,
            [0, 2],
        ),
        (
            " dhr\n  dhrwodn\n\ndhrwodn\n   dhrwodn",
            ["v", "/", "d", "h", "r", Qt.Key_Return, "n"],
            7,
            [0, 8],
        ),
        (
            " dhr\n  dhrwodn\n\ndhrwodn\n   dhrwodn",
            ["v", "/", "d", "h", "r", Qt.Key_Enter, "n", "N"],
            1,
            [0, 2],
        ),
    ],
)
def test_search_cmd_in_visual(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test / command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos == sel_pos_


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("a", ["v", "*"], 0, [0, 1]),
        ("a", ["v", "#"], 0, [0, 1]),
        ("dhr\n  Dhr\n\n_dhr\n   dhr", ["v", "*"], 19, [0, 20]),
        ("dhr\n  dhr\n\n_dhr\n   dhr", ["j2l", "v", "#"], 0, [0, 7]),
    ],
)
def test_asterisk_sharp_cmd_in_visual(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test *, # command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos == sel_pos_


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34", ["v", Qt.Key_Space], 1, [0, 2]),
    ],
)
def test_space_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test space command in visual."""
    _, _, editor, vim, qtbot = vim_bot

    CONF.set(CONF_SECTION, "leader_key", "F1")
    vim.apply_plugin_settings("")

    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34", ["l", "v", Qt.Key_Backspace], 0, [0, 2]),
    ],
)
def test_backspace_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test backspace command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34\n   k", ["v", Qt.Key_Enter], 9, [0, 10]),
    ],
)
def test_enter_cmd_in_v(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test enter command in visual."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


def test_jump_mark_in_visual(vim_bot):
    """Jump to mark while in visual mode."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(4)
    qtbot.keyClicks(cmd_line, "v'a")
    assert vim.vim_cmd.vim_status.vim_state == VimState.VISUAL
    assert editor.textCursor().position() == 0


@pytest.mark.parametrize(
    "start_line, cmd_list, expected",
    [
        (0, ["v", "]]"], 2),
        (6, ["v", "]]"], 6),
        (9, ["v", "]]"], 9),
    ],
)
def test_next_python_definition_in_visual(vim_bot, start_line, cmd_list, expected):
    """Jump to next Python definition while in Visual mode."""
    _, _, editor, vim, qtbot = vim_bot

    text = (
        "import os\n\n"
        "def a():\n    pass\n\n"
        "class Foo:\n    def b(self):\n        pass\n\n"
        "def c():\n    pass\n"
    )
    editor.set_text(text)

    block = editor.document().findBlockByNumber(start_line)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(block.position())
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().blockNumber() == expected


@pytest.mark.parametrize(
    "start_line, cmd_list, expected",
    [
        (10, ["v", "[["], 9),
        (6, ["v", "[["], 5),
        (2, ["v", "[["], 2),
    ],
)
def test_prev_python_definition_in_visual(vim_bot, start_line, cmd_list, expected):
    """Jump to previous Python definition while in Visual mode."""
    _, _, editor, vim, qtbot = vim_bot

    text = (
        "import os\n\n"
        "def a():\n    pass\n\n"
        "class Foo:\n    def b(self):\n        pass\n\n"
        "def c():\n    pass\n"
    )
    editor.set_text(text)

    block = editor.document().findBlockByNumber(start_line)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(block.position())
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().blockNumber() == expected


@pytest.mark.parametrize(
    "start_line, cmd_list, expected",
    [
        (0, ["v", "}}"], 3),
        (3, ["v", "}}"], 6),
        (6, ["v", "}}"], 9),
        (9, ["v", "}}"], 12),
        (12, ["v", "}}"], 17),
        (17, ["v", "}}"], 20),
    ],
)
def test_next_python_block_in_visual(vim_bot, start_line, cmd_list, expected):
    """Jump to next Python block while in Visual mode."""
    _, _, editor, vim, qtbot = vim_bot

    text = (
        "def a():\n"
        "    pass\n\n"
        "if x:\n    pass\n\n"
        "for i in range(3):\n    pass\n\n"
        "while False:\n    pass\n\n"
        "try:\n    pass\nexcept Exception:\n    pass\n\n"
        "with open('file') as f:\n    pass\n\n"
        "class Foo:\n    def bar(self):\n        pass\n"
    )
    editor.set_text(text)

    block = editor.document().findBlockByNumber(start_line)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(block.position())
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().blockNumber() == expected


@pytest.mark.parametrize(
    "start_line, cmd_list, expected",
    [
        (20, ["v", "{{"], 17),
        (17, ["v", "{{"], 12),
        (12, ["v", "{{"], 9),
        (9, ["v", "{{"], 6),
        (6, ["v", "{{"], 3),
        (3, ["v", "{{"], 0),
    ],
)
def test_prev_python_block_in_visual(vim_bot, start_line, cmd_list, expected):
    """Jump to previous Python block while in Visual mode."""
    _, _, editor, vim, qtbot = vim_bot

    text = (
        "def a():\n"
        "    pass\n\n"
        "if x:\n    pass\n\n"
        "for i in range(3):\n    pass\n\n"
        "while False:\n    pass\n\n"
        "try:\n    pass\nexcept Exception:\n    pass\n\n"
        "with open('file') as f:\n    pass\n\n"
        "class Foo:\n    def bar(self):\n        pass\n"
    )
    editor.set_text(text)

    block = editor.document().findBlockByNumber(start_line)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(block.position())
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().blockNumber() == expected



from __future__ import annotations

# Standard Libraries
import os.path as osp

# Third Party Libraries
from qtpy.QtCore import Qt
from qtpy.QtGui import QStandardItem

from .list_dialog import PopupTableDialog


class MarkListDialog(PopupTableDialog):
    """Dialog to display bookmarks and allow jumping to them."""

    _MIN_WIDTH = 800
    _MAX_HEIGHT = 600

    def __init__(self, marks: list[tuple[str, dict]], parent=None) -> None:
        super().__init__(
            "Marks",
            parent=parent,
            headers=["Mark", "Line", "Col", "File", "Text"],
            min_width=self._MIN_WIDTH,
            max_height=self._MAX_HEIGHT,
        )

        self.marks = marks
        self.selected_mark = ""

        self._populate()
        if self.list_model.rowCount() > 0:
            self.list_viewer.setCurrentIndex(self.list_model.index(0, 0))
            self.list_viewer.selectRow(0)

    # ------------------------------------------------------------------
    # Qt overrides
    # ------------------------------------------------------------------
    def accept(self) -> None:
        """Store selected mark and close the dialog."""
        row = self.list_viewer.currentIndex().row()
        if 0 <= row < len(self.marks):
            self.selected_mark = self.marks[row][0]
        else:
            self.selected_mark = ""
        super().accept()

    def reject(self) -> None:
        """Clear selected mark when dialog is cancelled."""
        self.selected_mark = ""
        super().reject()

    def _populate(self) -> None:
        self.list_model.setRowCount(0)
        for mark, info in self.marks:
            file_path = info.get("file", "")
            line = info.get("line", 0)
            col = info.get("col", 0)
            text = ""
            try:
                with open(file_path, "r", encoding="utf-8") as fh:
                    rows = fh.readlines()
                    if 0 <= line < len(rows):
                        text = rows[line].strip()
            except Exception:
                pass
            basename = osp.basename(file_path)
            row = [
                QStandardItem(mark),
                QStandardItem(str(line + 1)),
                QStandardItem(str(col + 1)),
                QStandardItem(basename),
                QStandardItem(text),
            ]
            for i, item in enumerate(row):
                item.setEditable(False)
                if i in (1, 2):
                    item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
            self.list_model.appendRow(row)

    def get_selected_mark(self) -> str:
        """Return the mark selected by the user."""
        return self.selected_mark



# -*- coding: utf-8 -*-
"""Tests for the cell helper utilities."""

# Third Party Libraries
from qtpy.QtGui import QTextDocument

# Project Libraries
from spyder_okvim.utils.cell_helpers import get_document_cells


def _document_end(document: QTextDocument) -> int:
    """Return the exclusive end position of ``document``."""
    last_block = document.lastBlock()
    return last_block.position() + last_block.length()


def _set_editor_text(editor, qtbot, text: str, expected_cells: int) -> None:
    """Set editor text and wait until cell list matches ``expected_cells``."""
    editor.set_text(text)
    editor.highlighter.rehighlight()

    def _cells_ready() -> bool:
        return len(editor.get_cell_list()) == expected_cells

    qtbot.waitUntil(_cells_ready, timeout=2000)


def test_get_document_cells_with_headers(vim_bot):
    """Ensure cell regions are extracted from explicit headers."""
    _, _, editor, _, qtbot = vim_bot

    original_text = editor.toPlainText()
    original_cells = len(editor.get_cell_list())

    try:
        text = (
            "# %% Cell A\n"
            "a = 1\n"
            "\n"
            "# %%\n"
            "b = 2\n"
            "# %%% Subcell\n"
            "c = 3\n"
        )
        _set_editor_text(editor, qtbot, text, expected_cells=3)

        regions = get_document_cells(editor)

        assert [region.name for region in regions] == [
            "Cell A",
            "Unnamed Cell",
            "Subcell",
        ]
        assert [region.start_line for region in regions] == [0, 3, 5]
        assert [region.level for region in regions] == [0, 0, 1]
        assert all(region.has_header for region in regions)
        assert regions[-1].end_line == editor.document().blockCount() - 1
    finally:
        _set_editor_text(editor, qtbot, original_text, expected_cells=original_cells)


def test_get_document_cells_without_headers(vim_bot):
    """The entire file is treated as a single region when no headers exist."""
    _, _, editor, _, qtbot = vim_bot

    original_text = editor.toPlainText()
    original_cells = len(editor.get_cell_list())

    try:
        text = "print('hello')\nprint('world')\n"
        _set_editor_text(editor, qtbot, text, expected_cells=0)

        regions = get_document_cells(editor)
        assert len(regions) == 1

        region = regions[0]
        assert region.name is None
        assert region.level == 0
        assert region.start_line == 0
        assert region.end_line == editor.document().blockCount() - 1
        assert region.start_position == 0
        assert region.end_position == _document_end(editor.document())
        assert region.has_header is False
    finally:
        _set_editor_text(editor, qtbot, original_text, expected_cells=original_cells)



# -*- coding: utf-8 -*-
"""Widgets used to display inline annotations."""

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QLabel


class InlineLabel(QLabel):
    """Small floating label shown over the editor text."""

    def __init__(self, parent=None):
        """Create the label widget."""
        super().__init__(parent)
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)

    def set_style(self, font_family: str, font_size_pt: int):
        """Apply styling for display.

        Args:
            font_family: Font family name.
            font_size_pt: Size of the font in points.
        """
        self.setStyleSheet(
            f"""QLabel {{
            background-color : #222b35;
            color : #cd4340;
            border: 1px solid #cd4340;
            padding : 0px;
            text-indent : 1px;
            margin : 0px;
            font-family: {font_family};
            font-size: {font_size_pt}pt;
            }}"""
        )







# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
"""Spyder okvim configuration page."""

# Third Party Libraries
from qtpy.QtCore import QRegularExpression, Qt
from qtpy.QtGui import QKeySequence, QRegularExpressionValidator
from qtpy.QtWidgets import QGridLayout, QGroupBox, QHBoxLayout, QLineEdit, QVBoxLayout
from spyder.api.preferences import PluginConfigPage


class OkvimConfigPage(PluginConfigPage):
    """Preferences page for configuring OkVim."""

    def setup_page(self):
        """Create configuration page for the plugin."""
        newcb = self.create_checkbox
        newce = self.create_coloredit
        newsb = self.create_spinbox
        newle = self.create_lineedit

        color_group = QGroupBox("Color")
        color_layout = QGridLayout()

        conf_color_info = {
            "Cursor Fg": ("cursor_fg_color", 0, 0),
            "Cursor Bg": ("cursor_bg_color", 0, 2),
            "Selection Fg": ("select_fg_color", 1, 0),
            "Selection Bg": ("select_bg_color", 1, 2),
            "Search Fg": ("search_fg_color", 2, 0),
            "Search Bg": ("search_bg_color", 2, 2),
            "Highlight after yank Fg": ("yank_fg_color", 3, 0),
            "Highlight after yank Bg": ("yank_bg_color", 3, 2),
        }

        for name, (option_name, idx_row, idx_col) in conf_color_info.items():
            label, clayout = newce(name, option_name, without_layout=True)
            label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
            clayout.lineedit.setValidator(
                QRegularExpressionValidator(QRegularExpression(r"#[0-9abcdefABCDEF]{6}"))
            )
            color_layout.addWidget(label, idx_row, idx_col)
            color_layout.addLayout(clayout, idx_row, idx_col + 1)
            color_group.setLayout(color_layout)

        options_group = QGroupBox("Options")
        options_layout = QVBoxLayout()
        options_layout.addWidget(newcb("ignorecase", "ignorecase"))
        options_layout.addWidget(newcb("smartcase", "smartcase"))
        options_layout.addWidget(newcb("use s for searching 2ch", "use_sneak"))

        hl_yank_layout = QHBoxLayout()
        hl_yank_layout.addWidget(newcb("highlight after yank", "highlight_yank"))
        hl_yank_layout.addWidget(
            newsb(
                "(duration ",
                "ms)",
                "highlight_yank_duration",
                min_=0,
                max_=2000,
                step=100,
            )
        )
        options_layout.addLayout(hl_yank_layout)

        options_group.setLayout(options_layout)

        leaderkey_group = QGroupBox("Leader Key Mapping")
        leaderkey_layout = QHBoxLayout()

        self.leaderkey_viewer = newle(
            "Leader key", "leader_key", alignment=Qt.Horizontal
        )
        self.leaderkey_viewer.textbox.setAlignment(Qt.AlignHCenter)
        self.leaderkey_viewer.textbox.setEnabled(False)

        self.leaderkey_edit = ShortcutLineEdit(self, self.leaderkey_viewer.textbox)

        leaderkey_layout.addWidget(self.leaderkey_viewer)
        leaderkey_layout.addWidget(self.leaderkey_edit)
        leaderkey_group.setLayout(leaderkey_layout)

        layout = QVBoxLayout()
        layout.addWidget(color_group)
        layout.addWidget(options_group)
        layout.addWidget(leaderkey_group)

        self.setLayout(layout)


class ShortcutLineEdit(QLineEdit):
    """QLineEdit that filters its key press and release events."""

    def __init__(self, parent, viewer):
        """Create line edit used to record a leader key sequence.

        Args:
            parent: Parent widget.
            viewer: Line edit that displays the captured shortcut.
        """
        super().__init__(parent)
        self.setPlaceholderText("Press leader key...")
        self.viewer = viewer

    def keyPressEvent(self, event):
        """Filter key presses to capture a valid shortcut.

        Args:
            event: Key event captured by Qt.
        """
        key = event.key()
        modifier = event.modifiers()

        if modifier != Qt.NoModifier:
            return
        if not key or key == Qt.Key_unknown:
            return

        key_str = QKeySequence(key).toString()
        self.viewer.setText(key_str)
